/*
  cmtzhash

  Experimental code for generating a hash table for reflections stored
  in an CMTZ data structure
*/

#include <string.h>
#include "cmtzlib.h"
#include "ccp4_parser.h"
#include "ccp4_general.h"
#include "cmtzman_lib.h"

/* Basic linked list data structure element */

/* Parameters for the parser */
enum {
  MAX_LINE_LEN = 200
};

/* Main program
   Build a hash table and characterise it
*/
int main(int argc, char **argv) {
  /* Declare variables */
  MTZ *mtz=NULL;
  MTZCOL *hkl[3];
  int bin,ind_xtal,ind_set,ind_col[3],ihkl[3],ihkl2[3];

  MtzReflHashTable* reftab;

  /* Keywords */
  int           ntok=0,itok,cont;
  char          line[MAX_LINE_LEN],*key;
  CCP4PARSERTOKEN *token=NULL;
  CCP4PARSERARRAY *parser;

  /* Binning array and stats */
  int binning[NHASH];
  float mean,sd;
  int sum,sum2,max,min,nnone,nbelow,nabove;

  /* General variables */
  int i;

  /* Process command line */
  if (ccp4fyp(argc,argv)) {
    ccperror(1,"Error parsing command line (see above)");
  }

  /* Initialise the binning array */
  for (i=0; i<NHASH; i++) binning[i] = 0;

  /* Read MTZ data */
  mtz = MtzGet("HKLIN",1);
  if (mtz == NULL)
    ccperror(1,"\nFATAL: failed to read file\n");

  /* Initialise and populate  the hash table */
  reftab = InitMtzReflHashTable(mtz);

  /* First we need to locate which columns in the MTZ are
     the indices */
  /* Locate hkl columns in each of the structures */
  MtzFindInd(mtz,&ind_xtal,&ind_set,ind_col);
  for (i=0; i<3; i++)
    /* hkl is an array of pointers to the Miller indices */
    hkl[i] = mtz->xtal[ind_xtal]->set[ind_set]->col[ind_col[i]];
  
  /* Now loop over all the reflections and calculate the hash
     for each one */
  for (i=0; i<MtzNref(mtz); i++) {
    /* Acquire the Miller indices for each reflection */
    GetHKLFromI(hkl,i,ihkl);
    bin = MTZhash(ihkl);
    binning[bin]++;
  }

  /* Calculate some basic stats */
  max = binning[0];
  min = max;
  nnone = 0;
  sum = 0.0;
  sum2 = 0.0;
  for (i=0; i<NHASH; i++) {
    if (binning[i] > max) max=binning[i];
    if (binning[i] < min) min=binning[i];
    if (binning[i] == 0) nnone++;
    sum += binning[i];
    sum2 += binning[i]*binning[i];
  }
  mean = ((float) sum)/((float) NHASH);
  sd = sqrt(((float) sum2)/((float) NHASH) - mean*mean);

  for (i=0; i<NHASH; i++) {
    if (binning[i] < mean) nbelow++;
    if (binning[i] > mean) nabove++;
  }

  printf("Hash table characteristics: NHASH = %d MULTIPLIER = %d\n",
	 NHASH,MULTIPLIER);
  printf("There are %d reflections\n",MtzNref(mtz));
  printf("Mean = %f SD = %f\n",mean,sd);
  printf("Max = %d Min = %d\n",max,min);
  printf("Below mean = %d above mean = %d\n",nbelow,nabove);
  printf("Number with zero = %d\n",nnone);

  /* Allow user to request H K L's and return index */  /* Set up the parser
     This is used to return the tokens and associated info
     Set maximum number of tokens per line to 20 */
  parser = (CCP4PARSERARRAY *) ccp4_parse_start(20);
  if (parser == NULL)
    ccperror (1,"Couldn't create parser array" );

  /* Set up handy shortcuts */
  key   = parser->keyword;
  token = parser->token;

  /* Read lines from stdin until END/end keyword is entered or
     EOF is reached */
  cont = 1;

  while (cont) {
    /* Blank the line before calling ccp4_parser
       to force reading from stdin */
    line[0] = '\0';

    /* Call ccp4_parser to read input line and break into tokens
       Returns the number of tokens, or zero for eof */
    ntok = ccp4_parser(line,MAX_LINE_LEN,parser,1);

    if (ntok < 1) {
      /* End of file encountered */
      cont = 0;
    } else {      
      /* Lookup keywords */
      if (ntok == 1 && ccp4_keymatch("END",key)) {
	/* END keyword */
	cont = 0;
      }
      if (ntok == 4 && ccp4_keymatch("HKL",key)) {
	/* Lookup a reflection index */
	ihkl[0] = (int) token[1].value;
	ihkl[1] = (int) token[2].value;
	ihkl[2] = (int) token[3].value;
	printf("Reflection: %d %d %d\n",ihkl[0],ihkl[1],ihkl[2]);
	i = LookupMtzRefl(reftab,ihkl,0,0);
	if (i > -1) {
	  printf("Index = %d\n",i);
	  GetHKLFromI(hkl,i,ihkl2);
	  printf("HKL = ( %d, %d, %d )\n",ihkl2[0],ihkl2[1],ihkl2[2]);
	} else {
	  printf("Not found\n");
	}
      }
    }
  }
  
  /* Dispose of MTZ data */
  if (mtz) MtzFree(mtz);

  /* Dispose of hash table */
  if (reftab) FreeMtzReflHashTable(reftab);

  /* Finish */
  ccperror(0, "Normal termination");
}
