/*
  CMTZMAN
  Copyright (C) 2005-6 Peter Briggs
  
  This code is distributed under the terms and conditions of the
  CCP4 licence agreement as `Part 2' (Annex 2) software.
  A copy of the CCP4 licence can be obtained by writing to the
  CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
 
  cmtzman.c
  
  Utility for editing MTZ crystal, dataset and column data
  including names and associations.
*/

/****************************************************************
 * Headers
 ****************************************************************/

#include <string.h>
#include <math.h>
#include "cmtzlib.h"
#include "ccp4_parser.h"
#include "ccp4_general.h"
#include "cmtzman_lib.h"

/****************************************************************
 * Define constants
 ****************************************************************/

/* Define numbers as constants not macros
   Kernighan and Pike, "The Practice of Programming" */

static char CMTZMAN_VERSION[] = "0.0.1";

/* Some limits */
enum {
  MAX_NAME_LEN = 200,
  MAX_LINE_LEN = 1000
};

/****************************************************************
 * Local function prototypes
 ****************************************************************/


/****************************************************************
 * Program code
 ****************************************************************/

static char rcsid[] = "$Id: cmtzman.c,v 1.12 2007/08/31 07:19:41 pjx Exp $";

/* Main program */

int main(int argc, char **argv) {

  /* Declarations */

  /* Reading MTZ file */
  MTZ *mtz=NULL,*import_mtz=NULL;
  MTZCOL *col,*col2;
  MTZSET *set,*set2;
  MTZXTAL *xtal,*xtal2;

  /* CCP4 parser parameters */
  int           ntok=0,itok,cont;
  char          line[MAX_LINE_LEN],import_hklin[MAX_LINE_LEN],*key;
  CCP4PARSERTOKEN *token=NULL;
  CCP4PARSERARRAY *parser;

  /* Other varibles */
  int i,j;
  float cell[6],wavelength;
  char name[MAX_NAME_LEN],new_name[MAX_NAME_LEN],attr[5],ctype[3];

  /****************
    PROGRAM START
  *****************/

  /* Perform CCP4 initialisations */
  ccp4ProgramName("cmtzman");
  ccp4RCSDate    ("$Date: 2007/08/31 07:19:41 $");
  ccp4_prog_vers (CMTZMAN_VERSION);
  ccp4_banner();
  if (ccp4fyp(argc,argv)) {
    ccperror(1,"Error parsing command line (see above)");
  }

  /**************************************************
   * Open and read in the MTZ data
   **************************************************/

  /* Open the principle MTZ file and read into memory */
  printf("\nOpening MTZ file for reading\n");
  WriteLogicalName(stdout,"HKLIN");
  mtz = MtzGet("HKLIN",1);
  if (mtz == NULL)
    ccperror(1,"\nFATAL: failed to read file\n");

  /* Check the file for batches */
  if (MtzNbat(mtz) > 0) {
    /* This program cannot handle batch MTZ files */
    printf("HKLIN is a batch MTZ file - cannot be edited with this program");
    ccperror (1,"\nFATAL: HKLIN is batch MTZ file\n");
  } 

  /********************
    Read the user input
  *********************/
  
  /* Set up the parser
     This is used to return the tokens and associated info
     Set maximum number of tokens per line to 20 */
  parser = (CCP4PARSERARRAY *) ccp4_parse_start(20);
  if (parser == NULL)
    ccperror (1,"Couldn't create parser array" );

  /* Set up handy shortcuts */
  key   = parser->keyword;
  token = parser->token;

  /* Read lines from stdin until END/end keyword is entered or
     EOF is reached */
  cont = 1;

  while (cont) {
    /* Blank the line before calling ccp4_parser
       to force reading from stdin */
    line[0] = '\0';

    /* Call ccp4_parser to read input line and break into tokens
       Returns the number of tokens, or zero for eof */
    ntok = ccp4_parser(line,MAX_LINE_LEN,parser,1);

    if (ntok < 1) {
      /* End of file encountered */
      cont = 0;
    } else {      
      /* Lookup keywords */

      /* DELETE keyword
	 Delete an existing crystal, dataset or column
	 Usage:
	 DELETE XTAL <xname>
	 DELETE DATASET <xname>/<dname>
	 DELETE COLUMN <xname>/<dname>/<label>
      */
      if (ccp4_keymatch("DELETE",key) == 1) {
	if (ntok != 3) {
	  /* Wrong number of arguments */
	  ccperror(1,"Wrong number of arguments for DELETE");
	} else {
	  if (ccp4_keymatch("XTAL",token[1].word)) {
	    /* Deleting a crystal */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    printf("Delete crystal \"%s\"\n",name);
	    /* Find the crystal */
	    xtal = MtzXtalLookup(mtz,name);
	    if (!xtal) {
	      ccperror(1,"Crystal not found");
	    }
	    printf("Located ok\n");
	    /* Delete the crystal */
	    if (!MtzDeleteXtal(mtz,xtal)) {
	      ccperror(1,"Error attempting to delete crystal");
	    }
	    printf("Successfully deleted crystal\n");
	  } else if (ccp4_keymatch("DATASET",token[1].word)) {
	    /* Deleting a dataset */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    printf("Delete dataset \"%s\"\n",name);
	    set = MtzSetLookup(mtz,name);
	    /* Find the dataset */
	    if (!set) {
	      ccperror(1,"Dataset not found");
	    }
	    printf("Located ok\n");
	    /* Delete the crystal */
	    if (!MtzDeleteDataset(mtz,set)) {
	      ccperror(1,"Error attempting to delete dataset");
	    }
	    printf("Successfully deleted dataset\n");
	  } else if (ccp4_keymatch("COLUMN",token[1].word)) {
	    /* Deleting a column */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    printf("Delete column \"%s\"\n",name);
	    /* Find the column */
	    col = MtzColLookup(mtz,name);
	    if (!col) {
	      ccperror(1,"Column not found");
	    }
	    printf("Located ok\n");
	    /* Delete the column */
	    if (!MtzDeleteColumn(mtz,col)) {
	      ccperror(1,"Error attempting to delete column");
	    }
	    printf("Successfully deleted column\n");
	  } else {
	    /* Unrecognised element */
	    ccperror(1,"Error in DELETE keyword: unrecognised element");
	  }
	}
	/* End of DELETE keyword */
      }
      
      /* MOVE keyword
	 Move an existing dataset or column
	 Usage:
	 MOVE DATASET <xname>/<dname> <xname_new>
	 MOVE COLUMN <xname>/<dname>/<label> <xname_new>/<dname_new>
      */
      if (ccp4_keymatch("MOVE",key) == 1) {
	if (ntok != 4) {
	  /* Wrong number of arguments */
	  ccperror(1,"Wrong number of arguments for MOVE");
	} else {
	  if (ccp4_keymatch("DATASET",token[1].word)) {
	    /* Moving a dataset */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("Move dataset \"%s\" to \"%s\"\n",name,new_name);
	    /* Find the dataset */
	    set = MtzSetLookup(mtz,name);
	    if (!set) {
	      ccperror(1,"Dataset not found");
	    }
	    printf("Located ok\n");
	    /* Find the target crystal */
	    xtal = MtzXtalLookup(mtz,new_name);
	    if (!xtal) {
	      ccperror(1,"Target crystal not found");
	    }
	    printf("Located target crystal ok\n");
	    /* Need a second check that the target crystal
	       doesn't already contain a dataset with this
	       name */
	    for (i=0; i<MtzNsetsInXtal(xtal); i++) {
	      set2 = MtzIsetInXtal(xtal,i);
	      if (strcmp(MtzDatasetName(set),MtzDatasetName(set2)) == 0) {
		/* Dataset names match */
		ccperror(1,"Dataset name already exists in target crystal");
	      }
	    }
	    /* Move the dataset */
	    if (!move_dataset(mtz,set,xtal)) {
	      ccperror(1,"Move failed");
	    }
	    printf("Moved ok\n");
	  } else if (ccp4_keymatch("COLUMN",token[1].word)) {
	    /* Moving a column */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("Move column \"%s\" to \"%s\"\n",name,new_name);
	    /* Find the column */
	    col = MtzColLookup(mtz,name);
	    if (!col) {
	      ccperror(1,"Column not found");
	    }
	    printf("Located ok\n");
	    /* Find the target dataset */
	    set = MtzSetLookup(mtz,new_name);
	    if (!set) {
	      ccperror(1,"Target dataset not found");
	    }
	    printf("Located ok\n");
	    /* Move the column */
	    if (!move_column(mtz,col,set)) {
	      ccperror(1,"Move failed");
	    }
	    printf("Moved ok\n");
	  } else {
	    /* Unrecognised element */
	    ccperror(1,"Error in MOVE keyword: unrecognised element");
	  }
	}
	/* End of MOVE keyword */
      }
      
      /* RENAME keyword
	 Rename an existing crystal, dataset or column
	 RENAME XTAL <xname> <xname_new>
	 RENAME DATASET <xname>/<dname> <dname_new>
	 RENAME COLUMN <xname>/<dname>/<label> <label_new>
      */
      if (ccp4_keymatch("RENAME",key) == 1) {
	if (ntok != 4) {
	  /* Wrong number of arguments */
	  ccperror(1,"Wrong number of arguments for RENAME");
	} else {
	  if (ccp4_keymatch("XTAL",token[1].word)) {
	    /* Renaming a crystal */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("Rename crystal \"%s\" to \"%s\"\n",name,new_name);
	    /* Find the source crystal */
	    xtal = MtzXtalLookup(mtz,name);
	    if (!xtal) {
	      ccperror(1,"Crystal not found");
	    }
	    printf("Located ok\n");
	    /* Find the target crystal */
	    xtal2 = MtzXtalLookup(mtz,new_name);
	    if (xtal2) {
	      ccperror(1,"Target crystal name already exists");
	    }
	    printf("Target not found: ok\n");
	    /* Rename the crystal */
	    MtzSetXtalName(xtal,new_name);
	  } else if (ccp4_keymatch("DATASET",token[1].word)) {
	    /* Renaming a dataset */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("Rename dataset \"%s\" to \"%s\"\n",name,new_name);
	    /* Find the source dataset */
	    set = MtzSetLookup(mtz,name);
	    if (!set) {
	      ccperror(1,"Dataset not found");
	    }
	    printf("Located ok\n");
	    /* Find the target dataset */
	    set2 = MtzSetLookup(mtz,new_name);
	    if (set2) {
	      ccperror(1,"Target dataset name already exists");
	    }
	    printf("Target not found: ok\n");
	    /* Rename the dataset */
	    MtzSetDatasetName(set,new_name);
	  } else if (ccp4_keymatch("COLUMN",token[1].word)) {
	    /* Renaming a column */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("Rename column \"%s\" to \"%s\"\n",name,new_name);
	    /* Find the source column */
	    col = MtzColLookup(mtz,name);
	    if (!col) {
	      ccperror(1,"Column not found");
	    }
	    printf("Located ok\n");
	    /* Find the target column */
	    col2 = MtzColLookup(mtz,new_name);
	    if (col2) {
	      ccperror(1,"Target column name already exists");
	    }
	    printf("Target not found: ok\n");
	    /* Rename the column */
	    MtzSetColName(col,new_name);
	  } else {
	    /* Unrecognised element */
	    ccperror(1,"Error in RENAME keyword: unrecognised element");
	  }
	}
	/* End of RENAME keyword */
      }

      /* NEW keyword
	 Make a new crystal or dataset
	 Usage:
	 NEW XTAL <xname> CELL <a> <b> <c> [<alpha> <beta> <gamma>]
	 NEW DATASET <xname> <dname> [WAVELENGTH <lamba>]
      */
      if (ccp4_keymatch("NEW",key)) {
	if (ntok < 3) {
	  /* Wrong number of arguments */
	  ccperror(1,"Wrong number of arguments for NEW");
	} else {
	  if (ccp4_keymatch("XTAL",token[1].word)) {
	    /* Making a new crystal */
	    if (!(ntok == 7 || ntok == 10)) {
	      ccperror(1,"Wrong number of arguments for NEW XTAL");
	    }
	    /* Extract the new crystal name */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    printf("New crystal \"%s\"\n",name);
	    /* Find the target crystal */
	    xtal = MtzXtalLookup(mtz,name);
	    if (xtal) {
	      ccperror(1,"Target crystal name already exists");
	    }
	    printf("Target not found: ok\n");
	    /* Extract the cell */
	    cell[3] = 90.0;
	    cell[4] = 90.0;
	    cell[5] = 90.0;
	    for (i=0; i<(ntok-4); i++) {
	      j = i+4;
	      if (token[j].isnumber) {
		cell[i] = (float) token[j].value;
	      } else {
		printf("Bad cell parameter at position %d: %s\n",
		       (j+1),token[j].fullstring);
		ccperror(1,"Bad cell parameter supplied for NEW XTAL");
	      }
	    }
	    /* Make a new crystal */
	    if (!MtzAddXtal(mtz,name,name,cell)) {
	      ccperror(1,"Failed to make new crystal");
	    }
	    printf("Successfully made new crystal\n");
	  } else if (ccp4_keymatch("DATASET",token[1].word)) {
	    /* Making a new dataset */
	    if (!(ntok == 4 || ntok == 6)) {
	      ccperror(1,"Wrong number of arguments for NEW DATASET");
	    }
	    /* Extract the names */
	    strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	    printf("Target crystal \"%s\"\n",name);
	    xtal = MtzXtalLookup(mtz,name);
	    if (!xtal) {
	      ccperror(1,"Target crystal name not found");
	    }
	    printf("Located target crystal ok\n");
	    /* Find the target dataset */
	    strncpy(new_name,token[3].fullstring,MAX_NAME_LEN);
	    printf("New dataset \"%s\"\n",new_name);
	    set = MtzSetLookup(mtz,new_name);
	    if (set) {
	      ccperror(1,"Target dataset name already exists");
	    }
	    printf("Target dataset not found: ok\n");
	    /* Extract the wavelength */
	    wavelength = 0.0;
	    if (ntok == 6) {
	      if (token[5].isnumber) {
		wavelength = (float) token[5].value;
	      } else {
		printf("Bad wavelength parameter at position 6: %s\n",
		       token[5].fullstring);
		ccperror(1,"Bad wavelength supplied for NEW DATASET");
	      }
	    }
	    /* Make a new dataset */
	    if (!MtzAddDataset(mtz,xtal,new_name,wavelength)) {
	      ccperror(1,"Failed to make new dataset");
	    }
	    printf("Successfully made new dataset\n");
	  } else {
	    /* Unrecognised */
	    ccperror(1,"Error in NEW keyword: unrecognised element");
	  }
	}
	/* End of NEW keyword */
      }
      
      /* IMPORT keyword
	 Merge the data from another MTZ file 
	 Usage: IMPORT <file2>.mtz */
      if (ccp4_keymatch("IMPORT",key) == 1) {
	if (ntok != 2) {
	  /* Wrong number of arguments */
	  ccperror(1,"Wrong number of arguments for IMPORT");
	} else {
	  /* Get the filename */
	  strncpy(import_hklin,token[1].fullstring,MAX_LINE_LEN);
	  /* Attempt to open this file */
	  printf("\nOpening MTZ file for import\n");
	  WriteLogicalName(stdout,import_hklin);
	  import_mtz = MtzGet(import_hklin,1);
	  if (import_mtz == NULL)
	    ccperror(1, "\nFATAL: failed to import file\n");
	  printf("\nOpened MTZ file\n");
	  /* Merge the files together */
	  if (! merge_mtzs(mtz,import_mtz)) {
	    ccperror(1,"Error importing MTZ file");
	  }
	  /* Release the memory */
	  if (import_mtz) {
	    MtzFree(import_mtz);
	    import_mtz = NULL;
	  }
	}
      }

      /* SET keyword
	 Sets attributes for crystals, datasets and columns
	 Usage:
	 SET XTAL <xname> CELL <a> <b> <c> [<alpha> <beta> <gamma>]
	 SET XTAL <xname> PROJECT <pname>
	 SET DATASET <dname> WAVELENGTH <lambda>
	 SET COLUMN <label> TYPE <ctype> 
      */
      if (ccp4_keymatch("SET",key) == 1) {
	if (ntok < 4) {
	  /* Wrong number of arguments */
	  ccperror(1,"Insufficient number of arguments for SET");
	} else {
	  /* Get the name of the data item */
	  strncpy(name,token[2].fullstring,MAX_NAME_LEN);
	  /* Get the name of the attribute being set */
	  strncpy(attr,token[3].word,5);
	  if (ccp4_keymatch("XTAL",token[1].word)) {
	    /* SET XTAL <xname> */
	    printf("Target crystal \"%s\"\n",name);
	    /* Find the target crystal */
	    xtal = MtzXtalLookup(mtz,name);
	    if (!xtal) {
	      ccperror(1,"SET XTAL: target crystal not found");
	    }
	    /* Determine the attribute that's being changed */
	    if (ccp4_keymatch("CELL",attr)) {
	      /* CELL update */
	      if (ntok == 7 || ntok == 10) {
		/* Extract the cell parameters */
		cell[3] = 90.0;
		cell[4] = 90.0;
		cell[5] = 90.0;
		for (i=0; i<(ntok-4); i++) {
		  j = i+4;
		  if (token[j].isnumber) {
		    cell[i] = (float) token[j].value;
		  } else {
		    printf("Bad cell parameter at position %d: %s\n",
			   (j+1),token[j].fullstring);
		    ccperror(1,"SET XTAL: bad cell parameter supplied");
		  }
		}
		/* Update the cell */
		if (MtzSetXtalCell(xtal,cell)) {
		  printf("Set cell to %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f\n",
			 cell[0],cell[1],cell[2],cell[3],cell[4],cell[5]);
		} else {
		  ccperror(1,"SET XTAL: failed to update cell");
		}
	      } else {
		/* Error in number of cell parameters */
		ccperror(1,"SET XTAL: wrong number of parameters for CELL");
	      }
	    } else if (ccp4_keymatch("PROJECT",attr)) {
	      /* PROJECT name update */
	      if (ntok == 5) {
		/* Extract new project name */
		strncpy(new_name,token[4].fullstring,MAX_NAME_LEN);
		MtzSetProjectName(xtal,new_name);
		printf("Updated project name to \"%s\"\n",new_name);
	      } else {
		ccperror(1,"SET XTAL: wrong number of parameters for PROJECT");
	      }
	    } else {
	      /* No idea */
	      ccperror(1,"SET: unrecognised crystal attribute");
	    }
	  } else if (ccp4_keymatch("DATASET",token[1].word)) {
	    /* SET DATASET <dname> */
	    printf("Target dataset \"%s\"\n",name);
	    set = MtzSetLookup(mtz,name);
	    if (!set) {
	      ccperror(1,"SET: target dataset not found");
	    }
	    /* Determine the attribute that's being changed */
	    if (ccp4_keymatch("WAVELENGTH",attr)) {
	      /* WAVELENGTH update */
	      if (ntok == 5) {
		/* Extract the wavelength */
		wavelength = 0.0;
		if (token[4].isnumber) {
		  wavelength = (float) token[4].value;
		  /* Update the wavelength */
		  if (MtzSetDatasetWavelength(set,wavelength)) {
		    printf("Updated wavelength to %6.2f\n",wavelength);
		  } else {
		    ccperror(1,"SET DATASET: failed to update wavelength");
		  } 
		} else {
		  printf("Bad wavelength parameter at position 6: %s\n",
			 token[4].fullstring);
		  ccperror(1,"SET DATASET: bad wavelength value supplied");
		}
	      } else {
		ccperror(1,"SET DATASET: wrong number of arguments");
	      }
	    } else {
	      /* No idea */
	      ccperror(1,"SET: unrecognised dataset attribute");
	    }
	  } else if (ccp4_keymatch("COLUMN",token[1].word)) {
	    /* SET COLUMN <label> */
	    printf("Target column \"%s\"\n",name);
	    col = MtzColLookup(mtz,name);
	    if (!col) {
	      ccperror(1,"SET COLUMN: target column not found");
	    }
	    /* Determine the attribute that's being changed */
	    if (ccp4_keymatch("TYPE",attr)) {
	      /* Column TYPE update */
	      if (ntok == 5) {
		/* Extract the type */
		strncpy(ctype,token[4].word,3);
		MtzSetColType(col,ctype);
		printf("Updated type to \"%s\"\n",ctype);
	      } else {
		ccperror(1,"SET COLUMN: wrong number of arguments for TYPE");
	      }
	    } else {
	      /* No idea */
	      ccperror(1,"SET: unrecognised column attribute");
	    }
	  } else {
	    /* Unrecognised data element */
	    ccperror(1,"SET: unrecognised data element");
	  }
	}
      }

      /* SHOW keyword
	 Prints the current MTZ header data to stdout */
      if (ccp4_keymatch("SHOW",key) == 1) {
	show_hierarchy(mtz);
      }
     
      /* END keyword
         Usage: END | EXIT */
      if ((ccp4_keymatch("END",key) || ccp4_keymatch("EXIT",key))
	  && parser->ntokens == 1) {
	printf("END of input\n");
        cont = 0;
      }
    }
      
    /* Loop around to read next line */
  }

  /* Finished processing */

  /* Write out the modified MTZ data structure
     to HKLOUT */
  printf("\nWriting MTZ file\n\n");
  /* Add a history line */
  /*
    MtzAddHistory(mtz,histline,1);
  */
  WriteLogicalName(stdout,"HKLOUT");
  MtzPut(mtz,"HKLOUT");

  /* Clean up parser array */
  ccp4_parse_end(parser);

  /* Free MTZ structure */
  if (mtz) MtzFree(mtz);

  /* Finish */
  ccperror(0, "Normal termination"); 
}


/****************************************************************
 * Local functions
 ****************************************************************/

