/*
     CMTZMAN_LIB
     Copyright (C) 2005-6 Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK
*/

/* cmtzman_lib.c

   Pseudo-library for CMTZMAN program

*/
#include "cmtzman_lib.h"

static char CMTZMAN_LIB_VERSION[] = "0.0.2";


/****************************************************************
  Hashing of MTZ reflection data
*****************************************************************/

/* The hash table functions provide a quick way to access the
   index of a reflection from its Miller indices. */

/* Create a new hash table */
MtzReflHashTable* InitMtzReflHashTable(MTZ *mtz) {

  MtzReflHashTable *refltab;
  MTZCOL *hkl[3];
  int i,ind_xtal,ind_set,ind_col[3],ihkl[3];

  /* Make an empty hash table */
  refltab = (MtzReflHashTable *) malloc(sizeof(MtzReflHashTable *));
  refltab->hashtable = (MtzHashedRefl **) malloc(sizeof(MtzHashedRefl *)*NHASH);
  for (i=0; i<NHASH; i++) refltab->hashtable[i] = NULL;

  /* Populate the hash table */

  /* Locate hkl columns in each of the structures */
  MtzFindInd(mtz,&ind_xtal,&ind_set,ind_col);
  for (i=0; i<3; i++)
    /* hkl is an array of pointers to the Miller indices */
    hkl[i] = mtz->xtal[ind_xtal]->set[ind_set]->col[ind_col[i]];

  /* Add reflections into the table */
  for (i=0; i<MtzNref(mtz); i++) {
    /* Acquire the Miller indices for each reflection */
    GetHKLFromI(hkl,i,ihkl);
    LookupMtzRefl(refltab,ihkl,1,i);
  }
  
  /* Return the table */
  return refltab;
}

/* Lookup reflection, with optional insert */
int LookupMtzRefl(MtzReflHashTable *refltab, int ihkl[3], int insert, int index) {
  int h;
  MtzHashedRefl *refl;

  h = MTZhash(ihkl);
  for (refl = refltab->hashtable[h]; refl != NULL; refl = refl->next) {
    if (refl->h == ihkl[0] && refl->k == ihkl[1] && refl->l == ihkl[2])
      return refl->index;
  }
  if (insert) {
    refl = (MtzHashedRefl *) malloc(sizeof(MtzHashedRefl));
    refl->h = ihkl[0];
    refl->k = ihkl[1];
    refl->l = ihkl[2];
    refl->index = index;
    refl->next = refltab->hashtable[h];
    refltab->hashtable[h] = refl;
    return index;
  }
  return -1;
}

/* Hash function for MTZ reflection indices */
int MTZhash(int ihkl[3]) {
  int h=0,i=0;
  for (i=0; i<3; i++) h = MULTIPLIER * h + ihkl[i];
  return abs(h % NHASH);
}

/* Release a hash table */
int FreeMtzReflHashTable(MtzReflHashTable* refltab) {
  MtzHashedRefl *refl,*next;
  int i;
  if (!refltab) {
    for (i=0; i<NHASH; i++) {
      for (refl = refltab->hashtable[i] ; refl != NULL; refl = next) {
	next = refl->next;
	free(refl);
      }
    }
  }
  free(refltab->hashtable);
  free(refltab);
  return 1;
}

/****************************************************************
  merge_mtzs
  Combine two MTZ structures
*****************************************************************/
int merge_mtzs(MTZ *mtz1, MTZ *mtz2) {

  /* Declarations */
  int i,j,k,p,pos,n=0,nreflns;
  int ind_xtal,ind_set,ind_col[3],ihkl[3];
  char xname2[65],dname2[65],label2[31];
  MTZXTAL *xtal2;
  MTZSET  *set2;
  MTZCOL  *col1,*col2,*new_cols[100],*old_cols[100],*hkl1[3],*hkl2[3];
  MtzReflHashTable* reftab;

  /* Check that we have something to work with */
  if (!mtz1 || !mtz2) {
    ccperror(2,"ERROR one or both MTZ structures are NULL");
    return 0;
  }

  /* Check that neither are batch MTZ files */
  if (MtzNbat(mtz1) > 0 || MtzNbat(mtz2) > 0) {
    ccperror(2,"ERROR one or both MTZ structures are batch files");
    return 0;
  }

  /* Check spacegroups */
  if (MtzSpacegroupNumber(mtz1) != MtzSpacegroupNumber(mtz2)) {
    ccperror(2,"ERROR spacegroups differ");
  }

  /* Loop over crystals in the second MTZ structure
     Make a new crystal for each in the first
     Check that the names are unique */
  for (i=0; i<MtzNxtal(mtz2); i++) {
    printf("Importing crystal %d\n",i);
    strncpy(xname2,mtz2->xtal[i]->xname,65);
    printf("\tName: \"%s\"\n",xname2);
    if (strcmp(xname2,"HKL_base") != 0) {
      /* Only work with non-HKL_base crystals */
      p = 48;
      if ((pos = strlen(xname2)) >= 65) {
	pos = 64;
      } else {
	pos;
	xname2[pos+1] = '\0';
      }
      while (p<58 && (MtzXtalLookup(mtz1,xname2) != NULL)) {
	/* Name is not unique
	   Append a digit and try again */
	xname2[pos] = (char) p;
	p++;
      }
      if (p == 58) {
	/* Failed to find a unique name */
	ccperror(2,"ERROR unable to generate unique crystal name");
	return 0;
      }
      printf("\tImporting as: \"%s\"\n",xname2);
      /* Make the new crystal */
      xtal2 = new_crystal(mtz1,xname2,mtz2->xtal[i]->cell);
      if (!xtal2) {
	ccperror(2,"ERROR unable to make the new crystal");
	return 0;
      }
      /* Loop over the datasets in the crystal being
	 imported */
      for (j=0; j<MtzNsetsInXtal(mtz2->xtal[i]); j++) {
	strncpy(dname2,mtz2->xtal[i]->set[j]->dname,65);
	p = 48;
	if ((pos = strlen(dname2)) >= 65) {
	  pos = 64;
	} else {
	  pos;
	  dname2[pos+1] = '\0';
	}
	while (p<58 && (MtzSetLookup(mtz1,dname2) != NULL)) {
	  /* Name is not unique
	     Append a digit and try again */
	  dname2[pos] = (char) p;
	  p++;
	}
	if (p == 58) {
	  /* Failed to find a unique name */
	  ccperror(2,"ERROR unable to generate unique dataset name");
	  return 0;
	}
	/* Make the new dataset */
	set2 = new_dataset(mtz1,dname2,xtal2,
			   mtz2->xtal[i]->set[j]->wavelength);
	if (!set2) {
	  ccperror(2,"ERROR unable to make the new dataset");
	  return 0;
	}
	/* Deal with each column in the dataset */
	for (k=0; k<MtzNcolsInSet(mtz2->xtal[i]->set[j]); k++) {
	  /*  */
	  strncpy(label2,mtz2->xtal[i]->set[j]->col[k]->label,31);
	  p = 48;
	  if ((pos = strlen(label2)) >= 31) {
	    pos = 30;
	  } else {
	    pos;
	    label2[pos+1] = '\0';
	  }
	  while (p<58 && (MtzColLookup(mtz1,label2) != NULL)) {
	    /* Name is not unique
	       Append a digit and try again */
	    label2[pos] = (char) p;
	    p++;
	  }
	  if (p == 58) {
	    /* Failed to find a unique name */
	    ccperror(2,"ERROR unable to generate unique column name");
	    return 0;
	  }
	  /* Make the new column
	     Intially the columns will contain no data so we will
	     populate them at the end */
	  if (!(col2 = 
		MtzAddColumn(mtz1,set2,label2,mtz2->xtal[i]->set[j]->col[k]->type))) {
	    ccperror(2,"ERROR unable to make the new column");
	    return 0;
	  }
	  /* Add to the list of columns */
	  new_cols[n] = col2;
	  old_cols[n] = mtz2->xtal[i]->set[j]->col[k];
	  n++;
	  /* End of loop over columns */
	}
	/* End of loop over datasets */
      }
    } else {
      printf("\t*** HKL_base will not be imported\n");
    }
    /* End of loop over crystals */
  }

  /* Locate hkl in each of the structures */
  MtzFindInd(mtz1,&ind_xtal,&ind_set,ind_col);
  for (i=0; i<3; i++)
    hkl1[i] = mtz1->xtal[ind_xtal]->set[ind_set]->col[ind_col[i]];
  MtzFindInd(mtz2,&ind_xtal,&ind_set,ind_col);
  for (i=0; i<3; i++)
    hkl2[i] = mtz2->xtal[ind_xtal]->set[ind_set]->col[ind_col[i]];

  /* Build a hash table for the target MTZ data structure */
  reftab = InitMtzReflHashTable(mtz1);

  /* Load the reflection data into the new columns */
  printf("Updating reflection data...\n");
  for (i=0 ; i<n; i++) {

    col2 = old_cols[i];
    col1 = new_cols[i];

    printf("\tImporting data for column \"%s\"\n",col1->label);

    nreflns = 0;
    for (j=0 ; j<MtzNref(mtz2); j++) {
      GetHKLFromI(hkl2,j,ihkl);
      /*k = GetIFromHKL(mtz1,hkl1,ihkl);*/
      k = LookupMtzRefl(reftab,ihkl,0,0);
      if (k < 0) {
	/* Need to append the new reflection */
	k = MtzAddRefln(mtz1,ihkl);
	LookupMtzRefl(reftab,ihkl,1,k);
      }
      col1->ref[k] = col2->ref[j];
      if (col1->min > col1->ref[k]) col1->min = col1->ref[k];
      if (col1->max > col1->ref[k]) col1->max = col1->ref[k];

      nreflns++;
      if (fmod(((double) nreflns),((double) 5000)) == 0.0) 
	printf("\t\t%d reflections imported\n",nreflns);
    }
  }
  printf("Completed\n");
  /* Return success */
  return 1;
}

/****************************************************************
  move_column
  Move a column inside the MTZ structure
*****************************************************************/
MTZCOL* move_column(MTZ *mtz, MTZCOL *col, MTZSET *new_set) {

  /* Declarations */
  MTZXTAL *xtal=NULL;

  /* Protect H K and L */
  if (strcmp(col->type,"H") == 0) {
    printf("You cannot reassign the Miller indices\n");
    return NULL;
  }

  /* Fetch the parent crystal from the structure */
  xtal = MtzSetXtal(mtz,new_set);
  if (!xtal) {
    ccperror(2,"Serious error: failed to get parent crystal");
    return NULL;
  }

  /* Reassign the column */
  MtzAssignColumn(mtz,col,MtzXtalName(xtal),MtzDatasetName(new_set));
  return col;
}

/****************************************************************
  new_crystal
  Create a new crystal in the MTZ structure
*****************************************************************/
MTZXTAL *new_crystal(MTZ *mtz, char *xname, float cell[]) {

  /* Declarations */
  MTZXTAL *xtal=NULL;

  /* Make the new crystal */
  if (! (xtal = MtzAddXtal(mtz,xname,xname,cell))) {
    printf("Failed to make a new crystal\n");
    return NULL;
  }

  /* Successful termination */
  printf("New crystal created\n");
  return xtal;
}

/****************************************************************
  new_dataset
  Create a new dataset in the MTZ structure
*****************************************************************/
MTZSET *new_dataset(MTZ *mtz, char *dname, MTZXTAL *xtal,
		    float wavelength) {

  /* Declarations */
  MTZSET  *set=NULL;

  /* Make the new dataset */
  if (! (set = MtzAddDataset(mtz,xtal,dname,wavelength))) {
    printf("Failed to make a new dataset\n");
    return NULL;
  }

  /* Successful termination */
  printf("New dataset created\n");
  return set;
}

/****************************************************************
  move_dataset
  Move a dataset inside the MTZ structure
  Return 1 on success, 0 on failure
*****************************************************************/
MTZSET *move_dataset(MTZ *mtz, MTZSET *set, MTZXTAL *new_xtal) {

  /* Declarations */
  int i,ncols;
  MTZSET *new_set=NULL;
  MTZCOL *col=NULL;

  /* Protect the base dataset */
  if (set->setid == 0) {
    printf("You cannot move the base dataset\n");
    return NULL;
  }

  /* Perform reassignment */
  /* Do this by making a new dataset with the same name and
     properties as the current dataset, and then reassigning
     the columns from one to the other */

  /* New dataset */
  if (! (new_set = new_dataset(mtz,MtzDatasetName(set),\
			       new_xtal,MtzDatasetWavelength(set)))) {
    printf("Failed to make new dataset\n");
    return NULL;
  }

  /* Loop over columns and reassign each one */
  ncols = MtzNcolsInSet(set);
  for (i = 0; i < ncols ; i++) {
    printf("Moving column %d:",(i+1));
    /* We don't use i here because each reassignment moves
       all the remaining columns down the list
       So we assign column zero each time */
    col = set->col[0];
    printf("\"%s\"\n",col->label);
    MtzAssignColumn(mtz,col,MtzXtalName(new_xtal),MtzDatasetName(new_set));
  }

  /* Remove the empty dataset */
  MtzDeleteDataset(mtz,set);
  return new_set;
}

/****************************************************************
 rename_attribute
 Change the name of the requested attribute
*****************************************************************/
int rename_attribute(MTZ *mtz, char *attribute, char *old,
		     char *new, char *parent) {

  /* Declarations */
  int  i,nproj;
  char tmpname[150];
  MTZXTAL *xtal=NULL;
  MTZSET  *set=NULL;

  /* Sort out the input */
  /* Check that the attribute is valid */
  if (strcmp(attribute,"CRYSTAL") != 0 &&
      strcmp(attribute,"DATASET") != 0 &&
      strcmp(attribute,"PROJECT") != 0) {
    printf("\nUnrecognised attribute \"%s\"\n",attribute);
    return 1;
  }

  /* Check that the new name is not invalid */
  if (fix_name(new,&tmpname[0]) == 1) {
    printf("New name \"%s\" is not a valid name\n",new);
    return 1;
  }

  /* Did the user specify a parent crystal? */
  if (strlen(parent) > 0) {
    xtal = MtzXtalLookup(mtz,parent);
    if (!xtal) {
      printf("Unable to locate parent crystal \"%s\"\n",parent);
      return 1;
    }
  }

  /* Lookup the crystal or dataset to be changed */
  if (strcmp(attribute,"CRYSTAL") == 0) {
    xtal = MtzXtalLookup(mtz,old);
    if (!xtal) {
      printf("Unable to locate crystal \"%s\"\n",old);
      return 1;
    }
    /* Check the uniqueness of the new name */
    if (MtzXtalLookup(mtz,new) != NULL) {
      printf("Crystal \"%s\" already exists\n",new);
      return 1;
    }
    /* Make the change */
    strncpy(xtal->xname,new,65);
    return 0;

  } else if (strcmp(attribute,"DATASET") == 0) {
    /* If a parent was supplied then make the full name */
    if (xtal) {
      sprintf(tmpname,"%s/%s",parent,old);
    } else {
      strncpy(tmpname,old,65);
    }
    /* Locate the dataset */
    set = MtzSetLookup(mtz,tmpname);
    if (!set) {
      printf("Unable to locate dataset \"%s\"\n",tmpname);
      return 1;
    }
    /* Check the uniqueness of the new name */
    if (xtal) {
      /* If a parent was supplied then make the full name */
      sprintf(tmpname,"%s/%s",parent,new);
    } else {
      /* Do reverse lookup to find the parent crystal */
      xtal = MtzSetXtal(mtz,set);
      sprintf(tmpname,"%s/%s",xtal->xname,new);
    }
    if (MtzSetLookup(mtz,tmpname) != NULL) {
      printf("Dataset \"%s\" already exists\n",tmpname);
      return 1;
    }
    /* Make the change */
    strncpy(set->dname,new,65);
    return 0;

  } else if (strcmp(attribute,"PROJECT") == 0) {
    /* If a parent was supplied then check that the current
       project name is correct */
    if (xtal) {
      if (strcmp(xtal->pname,old) != 0) {
	printf("Project for crystal \"%s\" doesn't match input\n",parent);
	return 1;
      }
      /* Make the change for this crystal only */
      printf("Changing project name for crystal \"%s\" only\n",parent);
      strncpy(xtal->pname,new,65);
      return 0;
    } else {
      /* If no parent was specified then update all occurances of the
	 project name */
      nproj = 0;
      for (i=0; i<MtzNxtal(mtz); i++) {
	if (strcmp(mtz->xtal[i]->pname,old) == 0) {
	  xtal = mtz->xtal[i];
	  printf("Updating project name for crystal \"%s\"\n",xtal->xname);
	  /* Make the change */
	  strncpy(xtal->pname,new,65);
	  nproj++;
	}
      }
      if (nproj == 0) {
	printf("No project names were updated\n");
	return 1;
      } else {
	return 0;
      }
    }
  }
  /* We should never reach this point */
  return 1;
}

/****************************************************************
 validate_names
 Check crystal, project and dataset names and fix them if
 they don't match the standard
*****************************************************************/
int validate_names(MTZ *mtz) {

  /* Declarations */
  int i,j,k;
  int invalid=0,updated;
  char name[101],fixed_name[65],c;
  MTZSET *set;
  MTZXTAL *xtal;

  printf("\n **** VALIDATE NAMES ****\n");

  /**************************************************
   * Loop over crystals, datasets and columns
   * and check each name
   **************************************************/

  for (i=0; i < mtz->nxtal; i++) {

    /* Crystal level */
    xtal = mtz->xtal[i];
    printf("\nExamining crystal %d: \"%s\" (project \"%s\")",
	   i+1,xtal->xname,xtal->pname);

    updated = 0;

    if (fix_name(xtal->xname,&fixed_name[0]) == 1) {
      invalid = 1;
      printf("\n          *** correcting crystal name: is now \"%s\"",
	     fixed_name);
      strncpy(xtal->xname,fixed_name,65);
      updated = 1;
    }

    /* Also check the project name */
    if (fix_name(xtal->pname,&fixed_name[0]) == 1) {
      invalid = 1;
      printf("\n          *** correcting project name: is now \"%s\"",
	     fixed_name);
      strncpy(xtal->pname,fixed_name,65);
    }

    if (!updated) printf(" ... ok");
    printf("\n");

    for (j=0; j < xtal->nset; j++) {

      /* Dataset level */
      set = xtal->set[j];
      printf("          dataset %d: \"%s\"",j+1,set->dname);

      if (fix_name(set->dname,&fixed_name[0]) == 1) {
	invalid = 1;
	printf("\n          \t*** correcting name: is now \"%s\"",
	       fixed_name);
	strncpy(set->dname,fixed_name,65);
      } else {
	printf(" ... ok");
      }
      printf("\n");
    
    }
  }
  /* Finished validate_names */
  return invalid;
}

/****************************************************************
 show_hierarchy
 Show crystal, project, dataset names and column data in memory
*****************************************************************/
int show_hierarchy(MTZ *mtz) {

  /* Declarations */
  int i,j,k;
  int nxtals,nsets,ncols;
  MTZXTAL *xtal;
  MTZSET *set;
  MTZCOL *col;

  nxtals = MtzNxtal(mtz);
  printf("\nNumber of crystals: %d\n\n",nxtals);
  for (i=0; i < nxtals; i++) {
    xtal = mtz->xtal[i];
    printf("Crystal: %s\n",xtal->xname);
    printf("Project: %s\n",xtal->pname);
    printf("Cell:");
    for (j=0; j < 6; j++) {
      printf(" %6.2f",xtal->cell[j]);
    }
    printf("\n");

    nsets = MtzNsetsInXtal(xtal);
    if (nsets > 0) {
      printf("\tNumber of datasets: %d\n\n",nsets);
      for (j=0; j < nsets; j++) {
	set = xtal->set[j];
	printf("\tDataset %d: %s\n",(j+1),set->dname);
	printf("\tWavelength: %8.2f\n",set->wavelength);

	ncols = MtzNcolsInSet(set);
	if (ncols > 0) {
	  printf("\tNumber of columns: %d\n",ncols);
	  for (k=0; k < ncols; k++) {
	    col = set->col[k];
	    printf("\t\t%s (type %s)\n",col->label,col->type);
	  }
	  printf("\n");
	} else {
	  printf("\tNo columns assigned to this dataset\n\n");
	}
      }
    } else {
      printf("\tNo datasets assigned to this crystal\n\n");
    }
  }

  return 1;
}

/****************************************************************
 * Supplementary functions
 ****************************************************************/

/*********************************************************
   Echo the logical name and associated file name
   to the requested output stream
   Return 1 if the logical name has a non-NULL associated
   value (filename) and 0 if not.
**********************************************************/
int WriteLogicalName(FILE *fp, char *logicalname)
{
  char *filename=NULL;
  if (getenv(logicalname) != NULL) {
    filename = strdup(getenv(logicalname));
    printf(" Logical name: %s   Filename: %s\n",logicalname,filename);
    return 1;
  } else {
    printf(" Logical name: %s   No associated filename\n",logicalname);
    return 0;
  }
}
  

/*********************************************************
   fix_name
   Check whether a crystal, project or dataset name is
   valid (i.e. only contains alphanumeric characters and
   is no longer than 64 characters long).
   Return a corrected version in "fixed_name".
   Return 1 if the name is not valid and needs to be
   corrected, and 0 if it is ok.
**********************************************************/
int fix_name(char *name, char *fixed_name) {
  int ii;
  char c;
  for (ii=0; ii < 65; ii++) {
    c = name[ii];
    if (c != '\0') {
      if (isalnum(c) || c == '_')
	fixed_name[ii] = c;
      else
	fixed_name[ii] = '_';
    } else {
      fixed_name[ii] = '\0';
      break;
    }
  }
  if (strcmp(name,fixed_name) != 0 || strlen(name) > 64) {
    return 1;
  } else {
    return 0;
  }
}

/*
  extract_names
  Extract crystal and dataset names from a string of the
  form "/<xname>/<dname>"
 */
int extract_names(char *fullname, char *xname, char *dname) {

  /* Declarations */
  int n;
  CCP4PARSERARRAY *parser;

  /* Initialise */
  xname[0] = '\0';
  dname[0] = '\0';

  /* Set up the parser
     This is used to return the tokens and associated info
     Set maximum number of tokens per line to 20 */
  parser = (CCP4PARSERARRAY *) ccp4_parse_start(20);
  if (parser == NULL)
    ccperror ( 1,"Couldn't create parser array" );

  /* Expect fullname to be of the form /x/y/z
     Reset delimiters and process the name */
  ccp4_parse_delimiters(parser,"/","/");
  n = ccp4_parse(fullname,parser);
  if (n > 0) {
    /* At least one element - take this to be xname */
    strncpy(xname,parser->token[0].fullstring,65);
    printf("1st component: %s\n",xname);
  }
  if (n > 1) {
    /* Second element - take this to be dname */
    strncpy(dname,parser->token[1].fullstring,65);
    printf("2nd component: %s\n",dname);
  }
  if (n > 2) {
    /* Extra elements ignored */
    printf("Trailing components of name ignored\n");
  } else if (n == 0) {
    /* Nothing extracted */
    printf("No components extracted\n");
  } 

 /* Clean up parser array */
  ccp4_parse_end(parser);
  return n;
}


/****************************************************************
  MTZ UTILITIES

  This are utility functions which should be considered for
  migration into the CMTZ library.
*****************************************************************/

char *MtzXtalName(MTZXTAL *xtal) {
  /* Return pointer to the MTZ xtal name */
  return xtal->xname;
}

int MtzSetXtalCell(MTZXTAL *xtal, float cell[]) {
  /* Update the cell parameters for this crystal */
  int i;
  for (i=0; i<6; i++) {
    xtal->cell[i] = cell[i];
  }
  return 1;
}

char *MtzSetProjectName(MTZXTAL *xtal, char *pname) {
  /* Update the project name for the crystal */
  return strncpy(xtal->pname,pname,65);
}

char *MtzDatasetName(MTZSET *set) {
  /* Return pointer to the MTZ dataset name */
  return set->dname;
}

float MtzDatasetWavelength(MTZSET *set) {
  /* Return wavelength for the MTZ dataset */
  return set->wavelength;
}

int MtzSetDatasetWavelength(MTZSET *set, float wavelength) {
  /* Set wavelength for the MTZ dataset */
  if (wavelength >= 0.0) {
    set->wavelength = wavelength;
    return 1;
  }
  return 0;
}

char *MtzSetXtalName(MTZXTAL *xtal, char *xname) {
  /* Update the MTZ xtal name */
  strncpy(xtal->xname,xname,65);
  return xtal->xname;
}

char *MtzSetDatasetName(MTZSET *set, char *dname) {
  /* Update the MTZ dataset name */
  strncpy(set->dname,dname,65);
  return set->dname;
}

char *MtzSetColName(MTZCOL *col, char *label) {
  /* Update the MTZ column label */
  strncpy(col->label,label,31);
  return col->label;
}

int MtzSetColType(MTZCOL *col, char *ctype) {
  /* Update the MTZ column type */
  strncpy(col->type,ctype,3);
  return 1;
}

/****************************************************************
  MtzAddRefln
  Add a new reflection (h k l) to an existing MTZ data structure

  This means adding an extra element to the end of each column
  It is likely that the MTZ structure will need to be resorted
  after performing these operations.
*****************************************************************/
int MtzAddRefln(MTZ *mtz, int ihkl[]) {

  /* Declarations */
  int i,j,k,nref;
  int ind_xtal,ind_set,ind_col[3];
  float ind;
  union float_uint_uchar uf;
  MTZCOL *col,*hkl[3];
  MTZXTAL *xtal;
  MTZSET *set;

  /*printf("MtzAddRefln: %d %d %d\n",ihkl[0],ihkl[1],ihkl[2]);*/

  /* Check for non-NULL pointer to MTZ */
  if (!mtz) return 0;

  /* Fetch pointers to the H K and L columns */
  MtzFindInd(mtz,&ind_xtal,&ind_set,ind_col);
  for (i=0; i<3; i++)
    hkl[i] = mtz->xtal[ind_xtal]->set[ind_set]->col[ind_col[i]];

  /* Get the current number of reflections */
  nref = MtzNref(mtz);

  /* Loop over all columns in the structure */
  for (i=0; i<MtzNxtal(mtz); i++) {
    xtal = mtz->xtal[i];
    for (j=0; j<MtzNsetsInXtal(xtal); j++) {
      set = xtal->set[j];
      for (k=0; k<MtzNcolsInSet(set); k++) {
	col = set->col[k];
	/* Resize the array */
	ccp4array_resize(col->ref,(nref+1));
	/* Set this to missing number flag
	   Code lifted from MtzAddColumn */
	if (strncmp (mtz->mnf.amnf,"NAN",3) == 0) {
	  uf = ccp4_nan();
	} else {
	  uf.f = mtz->mnf.fmnf;
	}
	col->ref[nref] = uf.f;
      }
    }
  }

  /* Set the h k l values and update the limits */
  for (i=0; i<3; i++) {
    ind = (float) ihkl[i]; 
    hkl[i]->ref[nref] = ind;
    if (hkl[i]->min > ind) hkl[i]->min = ind;
    if (hkl[i]->max < ind) hkl[i]->max = ind;
  }

  /* Update the number of reflections */
  mtz->nref++;

  /*puts("MtzAddRefln: done");*/

  /* Return the index of the new reflection */
  return mtz->nref-1;
}

/****************************************************************
 MtzDeleteXtal

 Remove a crystal from the MTZ structure
*****************************************************************/
int MtzDeleteXtal(MTZ *mtz, MTZXTAL *xtal) {

  int i,nxtals;

  if (!xtal) return 0;

  printf("Removing crystal \"%s\"\n",xtal->xname);
  
  /* Delete all datasets prior to trying to
     delete the parent crystal */
  if (MtzNsetsInXtal(xtal) > 0) {
    for (i=0; i<MtzNsetsInXtal(xtal); i++) {
      MtzDeleteDataset(mtz,xtal->set[i]);
    }
  }

  /* Free the array holding the column pointers */
  ccp4array_free(xtal->set);

  /* Find the matching pointer in the list of
     xtals for the parent MTZ */
  nxtals = MtzNxtal(mtz);
  for (i=0 ; i<nxtals ; i++) {
    if (xtal == mtz->xtal[i]) {
      printf("Located crystal at position %d\n",i);
      /* Move the last xtal to overwrite
	 the current one */
      mtz->xtal[i] = mtz->xtal[nxtals-1];
      --mtz->nxtal;
      break;
    }
  }

  /* Free the associated memory */
  free((void *) xtal);
  return 1;
}

/****************************************************************
 MtzDeleteDataset

 Remove a dataset from the MTZ structure
*****************************************************************/
int MtzDeleteDataset(MTZ *mtz, MTZSET *set) {

  int i,nsets;
  MTZXTAL *xtal=NULL;
  MTZCOL *col=NULL;

  if (!set) return 0;

  printf("Removing dataset \"%s\"\n",set->dname);

  /* Get parent crystal */
  xtal = MtzSetXtal(mtz,set);
  /* This is a serious error */
  if (!xtal) return 0;

  printf("\tParent crystal is \"%s\"\n",xtal->xname);
  
  /* Delete all columns prior to trying to
     delete the parent dataset */
  while (MtzNcolsInSet(set) > 0) {
    MtzDeleteColumn(mtz,set->col[0]);
  }

  /* Free the array holding the column pointers */
  ccp4array_free(set->col);

  /* Find the matching pointer in the list of
     datasets for the parent crystal */
  nsets = MtzNsetsInXtal(xtal);
  for (i=0 ; i<nsets ; i++) {
    if (set == xtal->set[i]) {
      printf("Located dataset at position %d\n",i);
      if (i != (nsets-1)) {
	/* Move the last dataset to overwrite
	   the current one */
	xtal->set[i] = xtal->set[nsets-1];
      }
      /* Decrement the number of sets */
      --xtal->nset;
      break;
    }
  }

  /* Free the associated memory */
  free((void *) set);
  return 1;
}

/****************************************************************
 MtzDeleteColumn

 Remove a column from the MTZ structure
*****************************************************************/
int MtzDeleteColumn(MTZ *mtz, MTZCOL *col) {

  int i,ncols;
  MTZSET *set=NULL;

  if (!col) return 0;

  printf("Removing column \"%s\"\n",col->label);
  
  /* Get the parent dataset pointer */
  set = MtzColSet(mtz,col);
  /* This is a serious error */
  if (!set) return 0;

  printf("\tParent dataset is \"%s\"\n",set->dname);

  /* Find the matching pointer in the list of
     columns for the parent dataset */
  ncols = MtzNcolsInSet(set);
  for (i=0 ; i<ncols ; i++) {
    if (col == set->col[i]) {
      printf("Located column at position %d\n",i);
      /* Release the memory associated with the
	 reflection data */
      if (set->col[i]) {
	MtzFreeCol(col);
      }
      if (i != (ncols-1) ) {
	/* Move the last column to overwrite
	   the current one */
	set->col[i] = set->col[ncols-1];
      }
      --set->ncol;
      break;
    }
  }
  /* Finished */
  return 1;
}

/***************************************************************
 GetHKLFromI

 Return the Miller indices for the i'th reflection in the file
****************************************************************/
int GetHKLFromI(MTZCOL *hkl[], int i, int ihkl[]) {
  int j;
  for (j=0; j<3; j++)
    ihkl[j] = (int) hkl[j]->ref[i];
  return 1;
}

/***************************************************************
 GetIFromHKL

 Given a set of Miller indices, locate its position in the file
****************************************************************/
int GetIFromHKL(MTZ *mtz, MTZCOL *hkl[], int ihkl[]) {
  int i;
  for (i=0; i<MtzNref(mtz); i++) {
    if ((int) hkl[0]->ref[i] == ihkl[0] &&
	(int) hkl[1]->ref[i] == ihkl[1] &&
	(int) hkl[2]->ref[i] == ihkl[2])
      return i;
  }
  return -1;
}
